/*************************************************************************
 * This file is part of input-overlay
 * github.con/univrsal/input-overlay
 * Copyright 2022 univrsal <uni@vrsal.xyz>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************/

#include "network.hpp"
#include "uiohook_helper.hpp"
#include "client_util.hpp"
#include <cstdio>

#include "gamepad_helper.hpp"

namespace network {
tcp_socket sock = nullptr;
netlib_socket_set set = nullptr;
buffer buf;
std::atomic<bool> network_loop;
bool connected = false;

std::thread network_thread;

bool start_connection()
{
    DEBUG_LOGN("Allocating socket... ");
    set = netlib_alloc_socket_set(1);

    if (!set) {
        printf(" netlib_alloc_socket_set failed: %s\n", netlib_get_error());
        return false;
    }
    printf("Done.\n");

    DEBUG_LOGN("Opening socket... ");

    sock = netlib_tcp_open(&util::cfg.ip);

    if (!sock) {
        printf("netlib_tcp_open failed: %s\n", netlib_get_error());
        return false;
    }

    printf("Done.\n");

    if (netlib_tcp_add_socket(set, sock) == -1) {
        DEBUG_LOG("netlib_tcp_add_socket failed: %s", netlib_get_error());
        return false;
    }

    DEBUG_LOG("Connection successful!");

    /* Send client name */
    if (!util::send_text(util::cfg.username)) {
        DEBUG_LOG("Failed to send username (%s): %s", util::cfg.username, netlib_get_error());
        return false;
    }
    network_loop = true;
    start_thread();
    connected = true;
    return true;
}

void start_thread()
{
    network_thread = std::thread(network_thread_method);
}

void network_thread_method()
{
    int listen_counter = 0;

    while (network_loop) {
        if (++listen_counter >= 10000) {
            listen_counter = 0;
            if (!listen()) /* Has a timeout of 1ms*/
            {
                DEBUG_LOG("Received quit signal");
                network_loop = false; // The rest will be taken care of in the main thread
                break;
            }
        }

        /* Copy buffered data from hooks */
        if (util::cfg.monitor_gamepad) {
            std::lock_guard<std::mutex> lock(libgamepad::buffer_mutex);
            if (libgamepad::buf.write_pos() > 0) {
                buf.write(libgamepad::buf.get(), libgamepad::buf.write_pos());
                libgamepad::buf.reset();
            }
        }

        if (util::cfg.monitor_keyboard || util::cfg.monitor_mouse) {
            std::lock_guard<std::mutex> lock(uiohook::buffer_mutex);
            if (uiohook::buf.write_pos() > 0) {
                buf.write(uiohook::buf.get(), uiohook::buf.write_pos());
                uiohook::buf.reset();
            }
        }

        /* Reset scroll wheel if no scroll event happened for a bit */
        if (uiohook::last_scroll_time > 0 && util::get_ticks() - uiohook::last_scroll_time >= SCROLL_TIMEOUT) {
            buf.write<uint8_t>(MSG_MOUSE_WHEEL_RESET);
            uiohook::last_scroll_time = 0;
        }

        /* Send any data written to the buffer */
        if (buf.write_pos() > 0) {
            if (!netlib_tcp_send(sock, buf.get(), buf.write_pos())) {
                DEBUG_LOG("netlib_tcp_send: %s", netlib_get_error());
                break;
            }
            buf.reset();
        }

        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }

    DEBUG_LOG("Network loop exited");
    if (util::cfg.monitor_keyboard || util::cfg.monitor_mouse) {
        // We have to stop the hook here since the main thread is blocked by uiohook
        uiohook::stop();
    }
}

int numready = 0;
bool listen()
{
    numready = netlib_check_socket_set(set, 0);

    if (numready == -1) {
        DEBUG_LOG("netlib_check_socket_set failed: %s", netlib_get_error());
        return false;
    }

    if (numready && netlib_socket_ready(sock)) {
        auto msg = util::recv_msg();

        switch (msg) {
        case MSG_NAME_NOT_UNIQUE:
            DEBUG_LOG("Nickname is already in use. Disconnecting...");
            return false;
        case MSG_NAME_INVALID:
            DEBUG_LOG("Nickname is not valid. Disconnecting...");
            return false;
        case MSG_SERVER_SHUTDOWN:
            DEBUG_LOG("Server is shutting down.");
            return false;
        case MSG_READ_ERROR:
            DEBUG_LOG("Couldn't read message.");
            return false;
        case MSG_REFRESH:;    /* fallthrough */
        case MSG_PING_CLIENT: /* NO-OP needed */
            return true;
        default:
        case MSG_INVALID:
            return false;
        }
    }
    return true;
}

bool init()
{
    if (netlib_init() == -1) {
        DEBUG_LOG("netlib_init failed: %s", netlib_get_error());
        return false;
    }
    return true;
}

void close()
{
    if (!network_loop)
        return;
    network_loop = false;

    /* Tell server we're disconnecting */
    if (connected) {
        buf.reset();
        buf.write<uint8_t>(MSG_CLIENT_DC);
        netlib_tcp_send(sock, buf.get(), buf.write_pos());
    }

    /* Give server time to process DC message */
    util::sleep_ms(100);
    network_thread.join();
    netlib_tcp_close(sock);
    netlib_quit();
}
}
